package cz.cuni.amis.pogamut.ut2004.lectures.tagbot;

import cz.cuni.amis.pogamut.base.utils.guice.AgentScoped;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbcommands.Initialize;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.ConfigChange;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.GameInfo;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.InitedMessage;
import cz.cuni.amis.pogamut.ut2004.lectures.tagbot.TagMap.Direction;
import cz.cuni.amis.pogamut.ut2004.tag.bot.UT2004BotTagController;
import cz.cuni.amis.pogamut.ut2004.utils.UT2004BotRunner;
import cz.cuni.amis.utils.exception.PogamutException;

/**
 * Simple bot that is meant to be used together with 'DM-TagMap'. It serves for the testing of {@link TagMap}. 
 * <p><p>
 * There is a 'map' directory containing 'DM-TagMap.ut2' that should be placed into 'UT2004/Maps' directory.
 * The map is simple "rectangle" with no internal walls / obstacles. Simple "rectangle" arena.
 * <p><p>
 * By observing the logs, you can estimate how fast the agent can run at 100% and some other constants
 * concerning running.
 *
 * @author Jakub Gemrot aka Jimmy
 */
@AgentScoped
public class RunAroundTagMapBot extends UT2004BotTagController {

	Direction running = Direction.NORTH; 
	
	@Override
    public Initialize getInitializeCommand() {
        return new Initialize().setName("RunAroundTagMapBot");
    }

    @SuppressWarnings("unchecked")
    @Override
    public void botInitialized(GameInfo gameInfo, ConfigChange config, InitedMessage init) {
    	info.getBotName().setInfo(running.name());
    }

    @Override
    public void logic() {
        log.info("--- Logic iteration ---");
        
        for (Direction direction : Direction.values()) {
        	log.info(direction + " distance: " + TagMap.getWallDistance(direction, info.getLocation()));
        }
        
        log.info("Running " + running + "...");
        log.info("Velocity " + info.getVelocity());        
        
        double distance = TagMap.getWallDistance(running, info.getLocation());
        
        if (distance < 200) {
        	running = running.clockwise();
        	log.info("SWITCHING TO " + running + " !");
        	info.getBotName().setInfo(running.name());
        }
        
        move.moveTo(info.getLocation().add(running.direction(400)));        
    }

    public static void main(String args[]) throws PogamutException {
    	// START 3 AGENTS FIRST
        new UT2004BotRunner(RunAroundTagMapBot.class, "RunAround").setMain(true).startAgent();        
    }
}
