package cz.cuni.amis.pogamut.ut2004.lectures.tagbot;

import java.util.logging.Level;

import cz.cuni.amis.pogamut.base.utils.guice.AgentScoped;
import cz.cuni.amis.pogamut.base.utils.math.DistanceUtils;
import cz.cuni.amis.pogamut.base3d.worldview.object.Location;
import cz.cuni.amis.pogamut.unreal.communication.messages.UnrealId;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbcommands.Dodge;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbcommands.Initialize;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.ConfigChange;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.ControlMessage;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.GameInfo;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.InitedMessage;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.Player;
import cz.cuni.amis.pogamut.ut2004.communication.messages.gbinfomessages.Self;
import cz.cuni.amis.pogamut.ut2004.factory.guice.remoteagent.UT2004ServerFactory;
import cz.cuni.amis.pogamut.ut2004.lectures.tagbot.TagMap.Direction;
import cz.cuni.amis.pogamut.ut2004.tag.bot.TagEvents;
import cz.cuni.amis.pogamut.ut2004.tag.bot.UT2004BotTagController;
import cz.cuni.amis.pogamut.ut2004.tag.protocol.messages.TagPlayerStatusChanged;
import cz.cuni.amis.pogamut.ut2004.tag.server.UT2004TagServer;
import cz.cuni.amis.pogamut.ut2004.tag.server.UT2004TagServerModule;
import cz.cuni.amis.pogamut.ut2004.utils.UT2004BotRunner;
import cz.cuni.amis.pogamut.ut2004.utils.UT2004ServerRunner;
import cz.cuni.amis.utils.IFilter;
import cz.cuni.amis.utils.collections.MyCollections;
import cz.cuni.amis.utils.exception.PogamutException;
import java.util.Collection;

/**
 * Example of the bot playing TAG-GAME according to {@link ControlMessage}s from {@link UT2004TagServer}.
 * <p><p> Tag! game is simple children play where they are chasing each other,
 * see http://en.wikipedia.org/wiki/Tag_%28game%29 <p><p> There is a 'map'
 * directory containing 'DM-TagMap.ut2' that should be placed into 'UT2004/Maps'
 * directory. The map is simple "rectangle" with no internal walls / obstacles.
 * Simple "rectangle" arena. <p><p> There is a class containing procedural
 * description of the DM-TagMap, (see {@link TagMap}), thus you can play with
 * low-level movement command when implementing Tag! game behavior. <p><p> Note
 * that the implementation is BLANK! You have to fill it for yourself.
 *
 * @author Jakub Gemrot aka Jimmy
 */
@AgentScoped
public class TagBot extends UT2004BotTagController {

    private static int number = 0;
    private TagEvents tagEvents;
    private UnrealId tagBotId;
    private Player playerToCatch = null;
    Direction running = Direction.NORTH;

    /**
     * Here we can modify initializing command for our bot.
     *
     * @return
     */
    @Override
    public Initialize getInitializeCommand() {
        return new Initialize().setName("Bohuslav_Machac" + (++number));
    }

    /**
     * The bot is initialized in the environment - a physical representation of
     * the bot is present in the game.
     *
     * @param config information about configuration
     * @param init information about configuration
     */
    @SuppressWarnings("unchecked")
    @Override
    public void botInitialized(GameInfo gameInfo, ConfigChange config, InitedMessage init) {
        tagEvents = new TagEvents(getWorldView()) {

            public void tagPlayerStatusChanged(TagPlayerStatusChanged event) {
                if (event.getTagStatus()) {
                    tagBotId = event.getBotId();
                    playerToCatch = null;
                }
            }
        };
        tagEvents.enableTagEvents();
    }

    /**
     * The bot is initialized in the environment - a physical representation of
     * the bot is present in the game.
     *
     * @param config information about configuration
     * @param init information about configuration
     */
    @Override
    public void botFirstSpawn(GameInfo gameInfo, ConfigChange config, InitedMessage init, Self self) {
    }

    /**
     * This method is called only once right before actual logic() method is
     * called for the first time.
     */
    @Override
    public void beforeFirstLogic() {
    }

    /**
     * Main method that controls the bot - makes decisions what to do next. It
     * is called iteratively by Pogamut engine every time a synchronous batch
     * from the environment is received. This is usually 4 times per second - it
     * is affected by visionTime variable, that can be adjusted in GameBots ini
     * file in UT2004/System folder.
     */
    @Override
    public void logic() {
        // mark that another logic iteration has began
        log.info("--- Logic iteration ---");

        if (!tag.isGameRunning()) {
            log.info("TAG GAME NOT RUNNING YET...");
            move.stopMovement();
            return;
        }

        Location myLocation = bot.getLocation();

        // FIRST OF ALL we have to check in what "high-level" state of the game we're in

        // player, who has the tag ... can be null at the beginning of the game
        Player tagged = players.getPlayer(tagBotId);

        if (tag.hasTag()) {
            // you HAVE THE TAG, chase other players ... do not chase players that are immune to you!
            log.info("I'm having the tag.");

            if (!players.canSeePlayers()) {
                move.turnHorizontal(150);
                return;
            }

            Collection<Player> candidates = MyCollections.getFiltered(players.getPlayers().values(), new IFilter<Player>() {

                @Override
                public boolean isAccepted(Player t) {
                    return !tag.isImmuneToMe(t);
                }
            });

            Player nearest = DistanceUtils.getNearest(candidates, myLocation);

            if (playerToCatch == null || nearest != playerToCatch) {
                playerToCatch = nearest;
            }

            if (!playerToCatch.isVisible()) {
                move.turnHorizontal(150);
                return;
            }

            double targetDistance = myLocation.getDistance(playerToCatch.getLocation());
            Location targetDirection = playerToCatch.getRotation().toLocation().getNormalized();

            if (targetDistance < 400) {
                move.turnTo(playerToCatch.getLocation().add(targetDirection.scale(targetDistance)));
                //move.dodge(new Location(1,0,0), true);
                act.act(new Dodge().setDirection(new Location(1, 0, 0)).setDouble(true));
            } else {
                move.moveTo(playerToCatch.getLocation().add(targetDirection.scale(targetDistance / 2.0)));

            }

        } else if (tag.isMeImmune(tagBotId)) {
            // you DO NOT HAVE the TAG and you are IMMUNE to the player who has the tag
            log.info("I'm immune to tag-holder.");
            if (tagged != null) {
                // player having the tag was seen
                if (tagged.isVisible()) {
                    // player having the tag is visible to you ~ you have accurate information about its location, rotation and velocity
                    log.info("I can see tag-holder.");
                    move.turnTo(tagged);

                    move.moveTo(myLocation.add(TagMap.getNearestWallDirection(tagged.getLocation()).opposite().direction().scale(500)));
                    //move.moveTo(tagged.getLocation().invert());


                } else {
                    // player having the tag IS NOT VISIBLE to you ~ you DO NOT HAVE accurate information about its location, rotation and velocity
                    log.info("I CANNOT see tag-holder.");
                    move.turnHorizontal(150);
                }


            } else {
                // player having the tag was not seen yet, try to look around
                log.info("I HAVE NOT SEEN the tag-holder YET.");
                move.turnHorizontal(150);
            }

        } else {
            // you DO NOT HAVE the TAG, you have to RUN AWAY from player who has the tag
            log.info("I can be tagged by tag-holder.");
            if (tagged != null) {
                // player having the tag was seen
                if (tagged.isVisible()) {
                    // player having the tag is visible to you ~ you have accurate information about its location, rotation and velocity
                    log.info("I can see tag-holder.");
                    move.turnTo(tagged);

                    if (myLocation.getDistance(tagged.getLocation()) < 350) {
                        move.dodge(TagMap.getNearestWallDirection(myLocation).opposite().direction(), true);
                        return;
                    }

                    if (TagMap.getNearestWallDistance(myLocation) <= 600) {
                        Direction nearestWall = TagMap.getNearestWallDirection(myLocation);
                        double clockwiseDist = TagMap.getWallDistance(nearestWall.clockwise(), myLocation);
                        double counterClockwiseDist = TagMap.getWallDistance(nearestWall.counterClockwise(), myLocation);
                        double diff = clockwiseDist - counterClockwiseDist;
                        if (diff > 0 && Math.random() < 0.5 + (0.5 / 1000.0) * diff) {
                            move.moveTo(myLocation.add(nearestWall.clockwise().direction().scale(500)));
                        } else {
                            move.moveTo(myLocation.add(nearestWall.counterClockwise().direction().scale(500)));
                        }

                    } else {

                        double scalar = info.getRotation().toLocation().dot2D(tagged.getRotation().toLocation());
                        if (scalar >= 0) {
                            move.moveTo(myLocation.sub(tagged.getLocation().getNormalized().scale(500)));
                        } else {
                            move.moveTo(myLocation.add(tagged.getRotation().toLocation().getNormalized().scale(500)));
                        }
                    }


                } else {
                    // player having the tag IS NOT VISIBLE to you ~ you DO NOT HAVE accurate information about its location, rotation and velocity
                    log.info("I CANNOT see tag-holder.");
                    move.turnHorizontal(150);
                }


            } else {
                // player having the tag was not seen yet, try to look around
                log.info("I HAVE NOT SEEN the tag-holder YET.");
                move.turnHorizontal(150);
            }
        }

    }

    public static void main(String args[]) throws PogamutException {
        // START 4 AGENTS FIRST
        new UT2004BotRunner(TagBot.class, "TagBot").setMain(false).setLogLevel(Level.WARNING).startAgents(1);

        // START UT2004TagServer
        //UT2004TagServerModule module = new UT2004TagServerModule();
        //UT2004ServerFactory factory = new UT2004ServerFactory(module);
        //UT2004ServerRunner serverRunner = new UT2004ServerRunner(factory);

        //UT2004TagServer server = (UT2004TagServer) serverRunner.setMain(false).startAgent();

        // START THE TAG GAME
        //server.startGame(5 * 60, 10);
    }
}
